// Watch.h

#ifndef _WATCH_h
#define _WATCH_h

#if defined(ARDUINO) && ARDUINO >= 100
	#include "arduino.h"
#else
	#include "WProgram.h"
#endif

#include <WROVER_KIT_LCD.h>
#include <Adafruit_GFX.h> //include TinyScreen library
#include <stdint.h>
#include <TimeLib.h> //include the Arduino Time library




typedef struct {
	bool  Second;
	bool  Minute;
	bool  Hour;
	bool  Wday;   // day of week, sunday is day 1
	bool  Day;
	bool  Month;
	bool  Year;   // offset from 1970;
}   TimeHasChanged_t;


enum
{
	
	clockBackgroundColor = ILI9341_DARKGREEN,

	clockCircleRadius = 50,

	clockMinuteHandLength = 40,
	clockMinuteHandColor = ILI9341_WHITE,

	clockHourHandLength = 35,
	clockHourHandColor = ILI9341_WHITE,


	clockSecondHandLength = 45,
	clockSecondHandColor = ILI9341_YELLOW,

	clockTextColor = ILI9341_ANTIQUEWHITE
};

enum
{
	clockPosHour		= 80,
	clockPosMin			= 100,
	clockPosSec			= 120,
	clockPosDay			= 180,
	clockPosMonth		= 200,
	clockPosYear		= 220,
	clockPosWDay		= 12,
	clockYFrame         = 13
};

class WatchClass
{
private:
	 Adafruit_GFX *_tft;
	 uint16_t _BackColor;
	 uint16_t _TextColor;
	 TimeHasChanged_t ChangedTime;
	 bool _FullMonth = true;

	uint16_t _ScreenHeight; 
	uint16_t _ScreenWidth;
	uint16_t _clockCenterX;
	uint16_t _clockCenterY;

	uint16_t _HourHand = clockBackgroundColor;
	uint16_t _MinHand = clockMinuteHandColor;
	uint16_t _SecHand = clockSecondHandColor;
	uint16_t _Background = clockBackgroundColor;

	 char Buffer[35];
	 void drawHand(int xStart, int yStart, int angle, int radius, uint16_t color);
 public:
	 tmElements_t sNewTime;
	 tmElements_t sOldTime;


	void init(Adafruit_GFX *tft, uint16_t BackColor, uint16_t TextColor,bool FullMonth,uint16_t x, uint16_t y);
	void initAnalogColors(uint16_t HourHand, uint16_t MinHand, uint16_t SecHand, uint16_t Background);
	void drawCircle(int x0, int y0, int radius, uint16_t color);
	
	void drawAdvancedCircle(void);
	void SetAnalogTime(tmElements_t Value);
	void ClearAnalogTime(tmElements_t Value);
	void SetDigiTime(tmElements_t OldValue, tmElements_t NewValue, bool OnOFF);
	
	TimeHasChanged_t SetTimeBool(tmElements_t NewValue, tmElements_t OldValue);

	


};

extern WatchClass Watch;

#endif

