







#include <Arduino.h>
#include "FS.h"

#include <SPIFFS.h>
#include <SPIFFSIniFile.h>
#include "soc/timer_group_struct.h"
#include "soc/timer_group_reg.h"

#include <SPI.h>
#include <Adafruit_GFX.h>    // Core graphics library
#include <Adafruit_ILI9341.h>
#include <Preferences.h>
#include <WiFi.h>
#include <MyTouch_XTP2046.h>
#include "Audio.h" // https://github.com/schreibfaul1/ESP32-audioI2S
#include <ArduinoOTA.h>

#include "Fonts/Picopixel.h"
#include "Fonts/comici8pt.c"
#include "Fonts/comici10pt.c"
#include "Fonts/Tiny3x3a2pt7b.h"
#include <ESPAsyncWebServer.h>  // 

#include "WiFiUdp.h"

#include <vector>

 #include "WiFi_config.inc"



Adafruit_ILI9341 tft = Adafruit_ILI9341(TFT_CS, TFT_DC, TFT_MOSI, TFT_CLK, TFT_RST, TFT_MISO);
TP tp(TOUCH_CS_PIN, TOUCH_IRQ_PIN);

#if AUDIO
Preferences pref;
Audio audio;
#endif

AsyncWebServer server(80);
AsyncWebSocket ws("/radio_ws");
AsyncWebSocketMessage * Message;
WiFiUDP Udp;

void audio_info(const char *info);
void TouchPad_released();


void draw_button(btns b);
void Write_StationNr(uint8_t nr);
void Write_Volume(uint8_t vol);
void write_stationName(String sName);
void write_streamTitle(String sTitle);

void SetStationByInt(int Value);
void SetVolume(int Value);

void AddNewStation(String url, String Name);
void DeleteStation(int pos);

void SetupButtons();
void DrawButtons();
void Init9341();
void Init9341XPT2046();
void InitAudio();
//void InitOTA();
void SetupPreference();

void connectToWiFi(const char * ssid, const char * pwd);
void WiFiEvent(system_event_id_t event);
void InitServer();
void InitOTA();
void InitSpiffs();
void ListSpiffs();
void SendStation2HTML(int Value);
String HTML_Processor(const String& var);
void onWinSocketEvent(AsyncWebSocket * server, AsyncWebSocketClient * client, AwsEventType type, void * arg, uint8_t *data, size_t len);
void ButtonWhatToDo(HTML_Bu_e Value);

void SendWSMessage(String Value);
HTML_Bu_e GotButton2Enum(String Value);
void FeedTheDog();

void SetHTMLStation();
int GetEntries(const char filename[]);
void SaveRadioStation_Name(String Value);
void SaveRadioStation(const char FileName[]);
void SetupStations(const char * filename);
void printErrorMessage(uint8_t e, bool eol = true);
String MakeIniStations();

Radio_t  GetRadioStation(int pos);

static inline __attribute__((always_inline)) float mapFloat( float x, const float in_min, const float in_max, const float out_min, const float out_max);
void SendUDPPackets(const char Value[] );
void udp_printf ( const char * format, ... );
void ReadFile(fs::FS &fs, const char * path);

//**************************************************************************************************
//                                           S E T U P                                             *
//**************************************************************************************************
void setup() {
  Serial.begin(115200);
  
  Init9341();
  Init9341XPT2046();
  SetupButtons();
  InitSpiffs();
  ListSpiffs();
  

  Serial.println("SetupStations(INI_FILENAME);");
  SetupStations(INI_FILENAME);

  Serial.println("SetHTMLStation();");
  SetHTMLStation(); 

  Serial.println("connectToWiFi(ssid, pass);");
  connectToWiFi(ssid, pass);
  
   InitServer();
   Serial.println("InitOTA();");
   InitOTA();


   InitAudio();
   SetupPreference();
  
  DrawButtons();

  Write_Volume(cur_volume);
  Write_StationNr(cur_station);
}
//**************************************************************************************************
//                                            L O O P                                              *
//**************************************************************************************************
void loop()
{
     FeedTheDog();

if (bWiFiConnected)
{	 
 
    if (millis() > _DelayNow + DelayPeriod)
    {
  // Serial.printf("--> %d ms - ", millis()) ;
 //  udp_printf("--> %d ms - ", millis()) ;

  _DelayNow = millis();

  if (bChangeSender)
  {
  bChangeSender = false; 
  SetStationByInt(cur_station);

  }
 
if (bSetVolume)
  {
  bSetVolume = false; 
  SetVolume(cur_volume);
  }

if ((bSetNewStationName) & (bSetNewStationURL))
{
  bSetNewStationName = false;
  bSetNewStationURL = false;    
  AddNewStation(_Radio.url, _Radio.Name);
  SaveRadioStation(INI_FILENAME);
  SetupStations(INI_FILENAME);
  ReadFile(SPIFFS, INI_FILENAME); // Just for Testing
  SetHTMLStation();
  SendWSMessage("[Refresh]");

    }

if (bDeleteStation)
{
   bDeleteStation = false;
   DeleteStation(cur_station);
   SaveRadioStation(INI_FILENAME);
   SetupStations(INI_FILENAME);
   ReadFile(SPIFFS, INI_FILENAME);
   SetHTMLStation();
   SendWSMessage("[Refresh]");
}


if (bStreamTitle)
{
bStreamTitle = false;
SendWSMessage("[Title]" + StreamTitle);
}
    }

       tp.loop();
        audio.loop();
           ArduinoOTA.handle();
           
  #if Debug 
  if (bWinSocketSend1)
{
 bWinSocketSend1 = false;
   
  sprintf(Buffer , "<- %d ->/0" , rand() );
}

   Serial.printf("Millies : %d" , millis());
   delay(250);
   #endif  
}

}
//**************************************************************************************************
//                                           E V E N T S                                           *
//**************************************************************************************************

/*********************************************************************************
* @brief  audio_info
* @param  char *info
* @return none 
*********************************************************************************/
void audio_info(const char *info)

{
  #if UDP_PRINT
  String sinfo = String(info);
  udp_printf("Audio info: %s \n", sinfo.c_str() );
  #else
    Serial.printf("Audio Info %s !! \n", info  );
  #endif  
}

/*********************************************************************************
* @brief  audio_showstation
* @param  char *info
* @return none 
*********************************************************************************/
void audio_showstation(const char *info)
{
    write_stationName(String(info));
  #if UDP_PRINT
    udp_printf("Auido Info %s !! \n", String(info).c_str()  );
  #else 
 Serial.printf("Audio Info %s !! \n", String(info).c_str()  );
   #endif   
}

/*********************************************************************************
* @brief  audio_showstreamtitle
* @param  char *info
* @return none 
*********************************************************************************/
void audio_showstreamtitle(const char *info)
{
    StreamTitle  = String(info);
    StreamTitle .replace("|", "\n");
    write_streamTitle( StreamTitle );
    #if UDP_PRINT
    udp_printf("Stream Title %s !! \n",  StreamTitle .c_str()  );
    #else
    Serial.printf("Stream Title %s !! \n",  StreamTitle .c_str()  );
    #endif

   bStreamTitle = true;

} 


/*********************************************************************************
* @brief  audio_showstreaminfo
* @param  char *info
* @return none 
*********************************************************************************/
void audio_showstreaminfo(const char *info)
{
String sinfo=String(info);
#if UDP_PRINT
udp_printf("audio_showstreaminfo %s !! \n", sinfo.c_str()  );
#else
Serial.printf("audio_showstreaminfo %s !! \n", sinfo.c_str()  );
#endif
}


/*********************************************************************************
* @brief  SetStationByInt
* @param  int Value
* @return none 
*********************************************************************************/
void SetStationByInt(int Value)
{
 Write_StationNr(Value);
 audio.connecttohost( GetRadioStation( Value ).url );
  pref.putShort("station", Value);
  SendStation2HTML(Value);
  ReadFile(SPIFFS, INI_FILENAME);
}


/*********************************************************************************
* @brief  SetVolume
* @param  int Value
* @return none 
*********************************************************************************/
void SetVolume(int Value)
{
    if (Value < 0)
    Value = 0;
    Write_Volume(Value);
    audio.setVolume(Value);
    pref.putShort("volume", Value);
}

/*********************************************************************************
* @brief  tp_released
* @param  None
* @return none 
*********************************************************************************/
void tp_released()
{
    
    Serial.printf("released Button : %d \n",  btn[cur_btn].a);
    
    #if AUDIO

    if(cur_btn !=-1){
        btn[cur_btn].s=RELEASED;
        draw_button(btn[cur_btn]);
        
        switch(btn[cur_btn].a){
            case VOLUME_UP:    cur_volume++;
                               if(cur_volume > max_volume)
                                cur_volume = max_volume;
                               
                                bSetVolume = true;
                                 // store the current volume in nvs
                              break;
           case VOLUME_DOWN:  cur_volume-- ; 
                              if(cur_volume < 0)
                                cur_volume = 0;
                                bSetVolume = true;
                                 // store the current volume in nvs
                              break;
            case STATION_UP:   cur_station++;
                                if(cur_station >  _Stations.Entries -1 )  
                                cur_station = 0; // then begin with Station 0

                                bChangeSender = true;
                                
                                #if UDP_PRINT
                                udp_printf((GetRadioStation( cur_station ).url).c_str() );
                                #else  
                                Serial.print(" ----> ");
                                Serial.println(GetRadioStation( cur_station ).url );
                                #endif
                                break;
           case STATION_DOWN : cur_station--;
                                if (cur_station < 0) // then Last Station
                                cur_station =  _Stations.Entries; 
                                bChangeSender = true;
                                #if UDP_PRINT
                                udp_printf((GetRadioStation( cur_station ).url).c_str() );
                                #else  
                                Serial.print(" -- current Station--> ");
                                Serial.print(cur_station);
                                Serial.println(GetRadioStation( cur_station ).url );
                                #endif
                                break;
        }
    }
    cur_btn=-1;

    #endif
}

/*********************************************************************************
* @brief  tp_pressed
* @param  X & Y Valur from Touch Pad
* @return none 
*********************************************************************************/
void tp_pressed(uint16_t X, uint16_t Y, uint16_t mx, uint16_t my)
{
 Serial.printf("tp pressed : %d - %d - %d - %d \n", X, X, mx, my);

    for(uint8_t i=0; i<(sizeof(btn)/sizeof(*btn)); i++)
    {
      #if UDP_PRINT
      udp_printf("Pressef X: %d -- Y: %d bx %d -- by %d \n", mx ,my, btn[i].y, btn[i].y+btn[i].h);
      #else
         Serial.printf("Pressef X: %d -- Y: %d bx %d -- by %d \n", mx ,my, btn[i].y, btn[i].y+btn[i].h);
   #endif

        if(mx>btn[i].x && (mx<btn[i].x+btn[i].w))
        {
      
        
            if(my>btn[i].y && (my<btn[i].y+btn[i].h))
            {
                cur_btn=i;
                btn[cur_btn].s=PRESSED;
                draw_button(btn[cur_btn]);
            }
        }
    }
}





/*********************************************************************************
* @brief  InitOTA()
* @param  none
* @return none
*********************************************************************************/
void InitOTA()
{
// OTA Configuration *************************************************

	ArduinoOTA.setPort(3232);
	// we use mDNS instead of IP of ESP32 directly 

	ArduinoOTA.setHostname("ESP32-Power-Consumption");

	// we set password for updating 
	ArduinoOTA.setPassword(OTA_PASS);
   #if UDP_PRINT 
      udp_printf(OTA_PASS);
    #else    
	   Serial.print(OTA_PASS);
    #endif
	// this callback function will be invoked when updating start 
	ArduinoOTA.onStart([]() {
    #if UDP_PRINT 
      udp_printf("Start updating");
     #else 
		  Serial.println("Start updating");
    #endif
	});
	// this callback function will be invoked when updating end 
	ArduinoOTA.onEnd([]() {
    #if UDP_PRINT 
    udp_printf("\nEnd updating");
    #else
 		Serial.println("\nEnd updating");
    #endif
	});
	// this callback function will be invoked when a number of chunks of software was flashed
	//so we can use it to calculate the progress of flashing 
	ArduinoOTA.onProgress([](unsigned int progress, unsigned int total) {
    #if UDP_PRINT 
    udp_printf("Progress: %u%%\r", (progress / (total / 100)));
    #else  
		Serial.printf("Progress: %u%%\r", (progress / (total / 100)));
    #endif
	});

	ArduinoOTA.onError([](ota_error_t error) {
    #if UDP_PRINT 
    udp_printf("Error[%u]: ", error);
    #else  
		Serial.printf("Error[%u]: ", error);
    #endif
		if (error == OTA_AUTH_ERROR) Serial.println("Auth Failed");
		else if (error == OTA_BEGIN_ERROR) Serial.println("Begin Failed");
		else if (error == OTA_CONNECT_ERROR) Serial.println("Connect Failed");
		else if (error == OTA_RECEIVE_ERROR) Serial.println("Receive Failed");
		else if (error == OTA_END_ERROR) Serial.println("End Failed");
	});
	// start updating 
	ArduinoOTA.begin();
  #if UDP_PRINT 
    udp_printf("****************************** Arduino Ota Start ***********************************");
     #else 
  Serial.println("****************************** Arduino Ota Start ***********************************");
    #endif
}

/*********************************************************************************
* @brief  InitServer()
* @param  none
* @return none
*********************************************************************************/
void InitServer()
{
  server.on("/heap", HTTP_GET, [](AsyncWebServerRequest *request){
    request->send(200, "text/plain", String(ESP.getFreeHeap()));
  });

  // Route for root / web page
  server.on("/", HTTP_GET, [](AsyncWebServerRequest *request){
    request->send(SPIFFS, "/index.html" , String(), false, HTML_Processor);
  });
  
  // Route to load style.css file
  server.on("/style.css", HTTP_GET, [](AsyncWebServerRequest *request){
    request->send(SPIFFS, "/style.css" , "text/css");
  });

  server.on("/java_script.js", HTTP_GET, [](AsyncWebServerRequest *request){
    request->send(SPIFFS, "/java_script.js" , "text/javascript");
  });

  server.on("/java_script2.js", HTTP_GET, [](AsyncWebServerRequest *request){
    request->send(SPIFFS, "/java_script2.js" , "text/javascript");
  });
 
  server.on("/audio0.jpg", HTTP_GET, [](AsyncWebServerRequest *request){
    request->send(SPIFFS, "/audio0.jpg", "image/jpg");
    });

  server.on("/audio1.png", HTTP_GET, [](AsyncWebServerRequest *request){
    request->send(SPIFFS, "/audio1.png", "image/png");
    });    

  server.on("/audio2.png", HTTP_GET, [](AsyncWebServerRequest *request){
    request->send(SPIFFS, "/audio2.png", "image/png");
    });    

  server.on("/audio3.png", HTTP_GET, [](AsyncWebServerRequest *request){
    request->send(SPIFFS, "/audio3.png", "image/png");
    });    

  server.on("/favicon.ico", HTTP_GET, [](AsyncWebServerRequest *request){
    request->send(SPIFFS, "/favicon.ico", "image/x-icon"); 
     }); 

 
/*
Just to memorize
const char* PARAM_INPUT_1 = "output";
const char* PARAM_INPUT_2 = "state";
const char* PARAM_INPUT_3 = "table";
*/
 
 ws.onEvent(onWinSocketEvent);
 server.addHandler(&ws);

  server.on("/buttons", HTTP_GET, [] (AsyncWebServerRequest *request) {
    String inputMessage1;
    if (request->hasParam(PARAM_INPUT_3))  {
      inputMessage1 = request->getParam(PARAM_INPUT_3)->value();
      }    else
    inputMessage1 = "No message sent";
   HTML_Bu = GotButton2Enum(inputMessage1);

       
   Serial.printf ("inputMessage1 - %s , %d ", inputMessage1.c_str(), HTML_Bu);
      
   ButtonWhatToDo(HTML_Bu);
    request->send(200, "/radio_ws","OK" ); 
  });

 server.begin();

}

/*********************************************************************************
* @brief  HTML_Processor (const String& var) For the HTML Webpage Values
* @param  var the Value to change
* @return String
*********************************************************************************/
String HTML_Processor(const String& var)
{
  String Status = "";
  Serial.print("----------- > ");
  Serial.println(var);
 
   if (var ==  "STATION_LIST")
   {
  Status = StationTabel; 
  Serial.print ("HTML Processesor ----------------- ");
   Serial.println(StationTabel);
   }
   else 
   if (var == "ESP_COMMAND")
   {
  Status =  ESP_COMMAND; 
  Serial.println("ESP_COMMAND");
  Serial.println(ESP_COMMAND);
   } else 
   if  (var == "IP_SERVER")
   Status = WiFi.localIP().toString();


 return (Status);
}



/*********************************************************************************
* @brief  InitSpiffs
* @param  none
* @return none 
*********************************************************************************/
void InitSpiffs()
{

// Initialize SPIFFS
  if(SPIFFS.begin(true))
    Serial.println("Mounting SPIFFS -SUCCESS !");
    else
    Serial.println("An Error has occurred while mounting SPIFFS");

    return;
}


/*********************************************************************************
* @brief  ListSpiffs List the file of ESP Flash
* @param  none
* @return none 
*********************************************************************************/
void ListSpiffs()
{
  
  File root = SPIFFS.open("/");
   File file = root.openNextFile();
 
  while(file){
 
      Serial.print("FILE: ");
      Serial.println(file.name());
      file = root.openNextFile();
  }
}


/*********************************************************************************
* @brief  SendWSMessage
* @param  String Value
* @return none 
*********************************************************************************/
void SendWSMessage(String Value)
 {
  ws.text( ClientID, Value.c_str() );
  ws.cleanupClients();
 }


/*********************************************************************************
* @brief  GotButton2Enum
* @param  Value 
* @return Enum 
*********************************************************************************/
HTML_Bu_e  GotButton2Enum(String Value)
 {
HTML_Bu_e Result = BuNone;

  for (int i = 0 ; i < BuCount ; i++)
   {
   if (HTML_Button[i] == Value){
   Result = (HTML_Bu_e)i;
  Value = "[Status]" + HTML_Button[i];
  SendWSMessage(Value);
  break;
}
   
   }  
return (Result);
 }


/*********************************************************************************
* @brief  SendStation2HTML
* @param  Value 
* @return none
*********************************************************************************/
void SendStation2HTML(int Value)
{
String Result;
itoa(Value, Buffer, 10);
Result = "[Station]"; 
Result += Buffer; 
ws.text( ClientID, Result.c_str() );
ws.cleanupClients();
}


//**************************************************************************************************
//                                              G U I                                              *
//**************************************************************************************************
/*********************************************************************************
* @brief  draw_button
* @param  btns b
* @return none
*********************************************************************************/
void draw_button(btns b)
{
    uint16_t ColorFrame = ILI9341_OLIVE;
    uint16_t ColorButton1 = ILI9341_CRIMSON;
    uint16_t ColorButton2 = ILI9341_CORAL;

    uint8_t r = 4, o = r * 3;
    if(b.s == RELEASED)
    { 
        ColorFrame = ILI9341_DARKRED;

    }
    if(b.s == PRESSED) 
    {
         ColorFrame = ILI9341_LIGHTSALMON;
    }

//    tft.fillRoundRect(b.x, b.y, b.w, b.h, r, ColorButton1);
    tft.fillRoundRect(b.x, b.y, b.w, b.h, r, ColorButton2);
    tft.drawRoundRect(b.x, b.y, b.w, b.h, r, ColorFrame );

     switch(b.a){
       case VOLUME_UP:
            tft.fillRoundRect(b.x, b.y, b.w, b.h, r, ColorButton2);
            tft.drawRoundRect(b.x, b.y, b.w, b.h, r, ColorFrame );
            tft.fillTriangle(b.x + b.w / 2, b.y + o, b.x + o, b.y +b.h - o , b.x + b.w - o , b.y + b.h - o, ILI9341_CRIMSON); break;
        case VOLUME_DOWN:
            tft.fillRoundRect(b.x, b.y, b.w, b.h, r, ColorButton2);
            tft.drawRoundRect(b.x, b.y, b.w, b.h, r, ColorFrame );
            tft.fillTriangle(b.x + o , b.y + o , b.x +b.w / 2, b.y + b.h - o, b.x + b.w - o, b.y + o, ILI9341_CRIMSON); break;
        case STATION_UP:
            tft.fillRoundRect(b.x, b.y, b.w, b.h, r, ColorButton1);
            tft.drawRoundRect(b.x, b.y, b.w, b.h, r, ColorFrame );
            tft.fillTriangle(b.x + b.w / 2, b.y + o, b.x + o, b.y +b.h - o , b.x + b.w - o , b.y + b.h - o, ILI9341_CORAL); break;
        case STATION_DOWN:
            tft.fillRoundRect(b.x, b.y, b.w, b.h, r, ColorButton1);
            tft.drawRoundRect(b.x, b.y, b.w, b.h, r, ColorFrame );
            tft.fillTriangle(b.x + o , b.y + o , b.x +b.w / 2, b.y + b.h - o, b.x + b.w - o, b.y + o, ILI9341_CORAL); break;
    }

/*
       case STATION_UP:
            tft.fillTriangle(b.x + o, b.y + o, b.x + o, b.y + b.h - o, b.x + b.w - o, b.y + b.h / 2 , color); break;
        case STATION_DOWN:
            tft.fillTriangle(b.x + b.w - o , b.y + o , b.x + o, b.y + b.h / 2 , b.x + b.w - o, b.y + b.h - o, color); break;
*/
}

    #define AjustY 50
    #define AjustX 25
    #define WRect 100
   

    #define HRect 65
    #define HRect2 40

    #define XRectA 10
    #define WRect2 ILI9341_TFTHEIGHT - XRectA * 2

    #define XRect1 40
    #define XRect2 190
    #define XRect3 300

    #define YRectA 5
    #define YRectB YRectA * 2 + HRect2
    #define YRect1 AjustY + HRect2 + YRectA * 2
    #define YRect2 300
    #define YRect3 200

void Write_StationNr(uint8_t nr)
{
    tft.fillRoundRect(XRect1 , YRect1,  WRect , HRect , 5, ILI9341_DARKGREEN);
    tft.drawRoundRect(XRect1 , YRect1 ,  WRect , HRect , 5, ILI9341_RED);
    String sNo = String(nr);
    if(sNo.length()<2) sNo = "0"+sNo;
    tft.setCursor(XRect1 + AjustX , YRect1 + HRect / 2 + 10);
    tft.setFont(&_comici8pt_);
    tft.setTextColor(ILI9341_LIGHTSALMON);
    tft.print(sNo);
    tft.setCursor(XRect1 + AjustX - 10, YRect1 + HRect / 2 + 20);
    tft.setFont(&Tiny3x3a2pt7b);
    tft.print("Station No");

}
void Write_Volume(uint8_t vol)
{
    tft.fillRoundRect(XRect2 , YRect1,  WRect , HRect , 5, ILI9341_DARKGREEN);
    tft.drawRoundRect(XRect2 , YRect1 ,  WRect , HRect , 5, ILI9341_RED);
    String svol = String(vol);
    if(svol.length()<2) svol = "0"+svol;
    tft.setCursor(XRect2 + AjustX , YRect1 + HRect / 2 + 4);
    tft.setFont(&_comici8pt_);
    tft.setTextColor(ILI9341_ALICEBLUE);
    tft.print(svol);
    tft.setCursor(XRect2 + AjustX , YRect1 + HRect / 2 +  20);
    tft.setFont(&Tiny3x3a2pt7b);
    tft.print("Volume");
}
void write_stationName(String sName)
{
    tft.fillRoundRect(XRectA,  YRectA,  WRect2 , HRect2, 5,  ILI9341_DARKOLIVEGREEN);
    tft.drawRoundRect(XRectA , YRectA , WRect2 , HRect2, 5, ILI9341_RED);
    tft.setFont(&Picopixel);
    tft.setTextColor(ILI9341_LIGHTGREY);
    tft.setCursor(XRectA + AjustX , YRectA + HRect2 / 2 );
    tft.print(sName);
}
void write_streamTitle(String sTitle)
{
    tft.fillRoundRect(XRectA, YRectB ,  WRect2 , HRect2, 5 , ILI9341_DARKOLIVEGREEN);
    tft.drawRoundRect(XRectA, YRectB ,  WRect2 , HRect2, 5, ILI9341_RED);
    tft.setFont(&Picopixel);
    tft.setTextColor(ILI9341_LIGHTGREY);
    tft.setCursor(XRectA + AjustX , YRectB +  HRect2 / 2 );
    tft.print(sTitle);
}


/*********************************************************************************
* @brief  SetupButtons
* @param  none
* @return none
*********************************************************************************/
void SetupButtons()
{
    #define YPos 180
    #define Height 46
    #define Width 54
    #define XFactor 80
    #define X1 10
    #define X2 X1 + XFactor 
    #define X3 X1 + XFactor * 2
    #define X4 X1 + XFactor * 3

    btn[0].x= X1; btn[0].y= YPos; btn[0].w=Width; btn[0].h=Height; btn[0].a=STATION_DOWN; btn[0].s=RELEASED;
    btn[1].x= X2; btn[1].y= YPos; btn[1].w=Width; btn[1].h=Height; btn[1].a=STATION_UP;   btn[1].s=RELEASED;
    btn[2].x= X3; btn[2].y= YPos; btn[2].w=Width; btn[2].h=Height; btn[2].a=VOLUME_UP;    btn[2].s=RELEASED;
    btn[3].x= X4; btn[3].y= YPos; btn[3].w=Width; btn[3].h=Height; btn[3].a=VOLUME_DOWN;  btn[3].s=RELEASED;
}

/*********************************************************************************
* @brief  DrawButtons
* @param  none
* @return none
*********************************************************************************/
void DrawButtons()
{
                                  
 Serial.println("Draw Buttons");       
for(uint8_t i=0;i<(sizeof(btn)/sizeof(*btn)); i++) 
{
draw_button(btn[i]);


                      #if UDP_PRINT
                      udp_printf("Button : x %d - y %d - h %d - w %d \n", btn[i].x , btn[i].y, btn[i].h, btn[i].w  );
                      #else  
                      Serial.printf("Button : x %d - y %d - h %d - w %d \n", btn[i].x , btn[i].y, btn[i].h, btn[i].w  );
                      #endif

}
}


/*********************************************************************************
* @brief  Init9341
* @param  none
* @return none
*********************************************************************************/
void Init9341()
{

Serial.println("Init9341");
	tft.begin();

	pinMode(TFT_LED, OUTPUT);
	digitalWrite(TFT_LED, HIGH);

	// read diagnostics (optional but can help debug problems)
	uint8_t x = tft.readcommand8(ILI9341_RDMODE);
	Serial.print("Display Power Mode: 0x"); Serial.println(x, HEX);
	x = tft.readcommand8(ILI9341_RDMADCTL);
	Serial.print("MADCTL Mode: 0x"); Serial.println(x, HEX);
	x = tft.readcommand8(ILI9341_RDPIXFMT);
	Serial.print("Pixel Format: 0x"); Serial.println(x, HEX);
	x = tft.readcommand8(ILI9341_RDIMGFMT);
	Serial.print("Image Format: 0x"); Serial.println(x, HEX);
	x = tft.readcommand8(ILI9341_RDSELFDIAG);
	Serial.print("Self Diagnostic: 0x"); Serial.println(x, HEX);
	tft.setRotation(TFT_NORMAL); // Set ILI9341 to Landscape

	tft.fillScreen(BgrColor);
	// End of Init ILI9341
	tft.setTextColor(ILI9341_ORANGE, ILI9341_NAVY);
  tft.setTextSize(2);
	tft.setTextWrap(false);
  tft.setCursor(10,10);
  tft.println("### Hallo I2S  ###");
  tft.println("### Internet Radio  ###");
  delay(2500);
  tft.fillScreen(BgrColor);
  
}

/*********************************************************************************
* @brief  Init9341XPT2046
* @param  none
* @return none
*********************************************************************************/
void Init9341XPT2046()
{

  SPI.begin(TOUCH_CLK, TOUCH_DIN, TOUCH_OUT );
  SPI.setFrequency( 60000000 );
  tp.setRotation(2);
  
}

/*********************************************************************************
* @brief  InitAudio
* @param  none
* @return none
*********************************************************************************/
void InitAudio()
{
   #if AUDIO
    Serial.println("Init Audio");
    audio.setPinout(I2S_BCLK, I2S_LRC, I2S_DOUT);
    audio.setVolume(cur_volume); // 0...21
    audio.connecttohost(  GetRadioStation( cur_station ).url );
  #endif 
}

/*********************************************************************************
* @brief  SetupPreference
* @param  none
* @return none
*********************************************************************************/
void SetupPreference()
{
   #if AUDIO  
    Serial.println("SetupPreference");
       pref.begin("WebRadio", false);  // instance of preferences for defaults (station, volume ...)

      if(pref.getShort("volume", 1000) == 1000){ // if that: pref was never been initialized
        Serial.printf("------------------ put Short -----------------------------------");
        pref.putShort("volume", 3);
        pref.putShort("station", 1);

    }
    else{ // get the stored values
    Serial.printf("******************************* get Short *******************************************");
        cur_station = pref.getShort("station");
        cur_volume = pref.getShort("volume");
    }

    SetStationByInt(cur_station);
	  SetVolume(cur_volume);

   #endif
}



/*********************************************************************************
* @brief connectToWiFi
* @param  none
* @return none
*********************************************************************************/
void connectToWiFi(const char * ssid, const char * pwd)
{
  	 Serial.println("Connect to WiFi");

     Serial.printf("SSID : %s - Password : %s ", ssid, pwd );
      
        // Configures static IP address
  if (!WiFi.config(local_IP, gateway, subnet, primaryDNS, secondaryDNS)) {
    Serial.println("STA Failed to configure");
  }
      
      Serial.printf(Buffer, "Connecting to WiFi network : %s \n", ssid);

  // delete old config
  WiFi.disconnect(true);
  WiFi.setHostname("ESP32_I2S_Radio");
  WiFi.onEvent(WiFiEvent);
  //register event handler
 // WiFi.onEvent(WiFiEvent);
  //Initiate connection
  WiFi.begin(ssid, pwd);
    while (WiFi.status() == bWiFiConnected) {
    delay(500);
    Serial.printf("- %d -", bWiFiConnected );
  }
  // Print local IP address and start web server

  #if UDP_PRINT
  udp_printf("WiFi connected!! --  IP address: %s", WiFi.localIP().toString().c_str());
  #else  
  Serial.print("<--------------- WiFi connected with - ");
  Serial.print("IP address: ");
  Serial.println(WiFi.localIP());
   #endif  
  
   }

/*********************************************************************************
* @brief WiFiEvent(system_event_id_t event)
* @param  WiFi Event
* @return none
*********************************************************************************/
void WiFiEvent(system_event_id_t event)
{
	Serial.print("event ");
Serial.print(event);

switch (event)
{
case     SYSTEM_EVENT_WIFI_READY:
 Serial.print("Wifi Ready");
	break;
case     SYSTEM_EVENT_STA_START:
 Serial.print("Wifi Event Start");
	break;
case     SYSTEM_EVENT_STA_STOP:
 Serial.print("Wifi Event Stop");
	break;
case     SYSTEM_EVENT_STA_CONNECTED:
{
	
  Serial.print("Wifi Connected");
}
	break;
case     SYSTEM_EVENT_STA_DISCONNECTED:
{
bWiFiConnected = false;
Serial.print("Wifi DisConnected");
}

   break;
case     SYSTEM_EVENT_STA_GOT_IP:
Serial.print("Got IP Adress : ");
 Serial.println(WiFi.localIP());
 Udp.begin(WiFi.localIP(), UdpPort);
 strcpy(Buffer, "Got IP Address");
  bWiFiConnected = true;

	break;
case     SYSTEM_EVENT_STA_LOST_IP:
 Serial.print("Wifi lost IP");
 bWiFiConnected = false;
	break;
 default:
 Serial.print("Not recognized Wifi Event ");
}}

static inline __attribute__((always_inline)) float mapFloat( float x, const float in_min, const float in_max, const float out_min, const float out_max)
{
  return ( x - in_min ) * ( out_max - out_min ) / ( in_max - in_min ) + out_min;
}


/*********************************************************************************
* @brief onWinSocketEvent
* @param  AsyncWebSocket * server, AsyncWebSocketClient * client, AwsEventType type, void * arg, uint8_t *data, size_t len
* @return none
*********************************************************************************/
void onWinSocketEvent(AsyncWebSocket * server, AsyncWebSocketClient * client, AwsEventType type, void * arg, uint8_t *data, size_t len)
{

static bool bStationURL  = false;
static bool bStationName = false;

String sValue;

ClientID = client->id();

  if(type == WS_EVT_CONNECT){
 
    Serial.println("Websocket client connection received");
 
  } else if(type == WS_EVT_DISCONNECT){
    Serial.println("Client disconnected");
 
  } else if(type == WS_EVT_DATA){
 
    Serial.println("Data received: ");
 
    for(int i=0; i < len; i++) 
      Buffer[i] = (char) data[i];
      
       Buffer[len] = '\0';
       Serial.println(Buffer);
       sValue  = Buffer;

     if (sValue.startsWith("[MsgName]"))
     {
       bStationName = true;
      int pos = sValue.indexOf('=');
       sValue.remove(0,pos +1);
      _Radio.Name = sValue;  

       Serial.print(" Name - ");
       Serial.println(_Radio.Name.c_str());
       bSetNewStationName = true;
            }

    if (sValue.startsWith("[MsgURL]"))
     {
       bStationURL = true;
      int pos = sValue.indexOf('=');
       sValue.remove(0,pos +1);
       _Radio.url = sValue;  
       Serial.print(" URL - ");
       Serial.println(_Radio.url.c_str());
       bSetNewStationURL = true;
            }     


        if (sValue.startsWith("[MsgSelect]"))
     {
       
      int pos = sValue.indexOf('=');
       sValue.remove(0,pos +1);

       sValue.trim();
       cur_station =  sValue.toInt(); 
       bChangeSender = true;
         
            }     
     
      client->printf("[Command] Command %s" , sValue.c_str() );
      if (bStationURL && bStationName)
      {
       sValue = _Radio.url + "- " + _Radio.Name;
       client->printf("[Status]  %s" , sValue.c_str() );
       bStationURL  = false;
       bStationName = false;
              
      }
  }
}


/*********************************************************************************
* @brief ButtonWhatToDo 
* @param  HTML_Bu_e Value as enum
* @return none
*********************************************************************************/
void ButtonWhatToDo(HTML_Bu_e Value)
{

switch (Value)
{

case  BuPrev:
      cur_station--;

      if (cur_station < 0) // then Last Station
          cur_station =  _Stations.Entries -1; 

      bChangeSender = true;
      Serial.println("BuPrev");
      break;
case  BuNext: 
      cur_station++;
      Serial.printf("1.) Die Station ist >%d< - max Entries %d \n", cur_station, _Stations.Entries);      
  
      if(cur_station >  _Stations.Entries )  
      cur_station = 0; // then begin with Station 0
		  Serial.printf("2.) Die Station ist >%d< - max Entries %d \n", cur_station, _Stations.Entries);       
      
      bChangeSender = true;
      Serial.println("BuNext");
      break;
case  BuVol_m: 
      if  (cur_volume > 0 )
      {
      cur_volume-- ;
      bSetVolume = true;
      Serial.println("BuVol_m");
      }
      break;
case  BuVol_p:
      if  (cur_volume < max_volume )
      cur_volume++ ;
      bSetVolume = true;
      Serial.println("BuVol_p");
      break;

case  BuMute: 
      audio.pauseResume();  
      break;

case  BuStatus: break;

case BuDelete:  
     bDeleteStation = true;
     break;

default: break;
}

}


/*********************************************************************************
* @brief FeedTheDog() Prevent a Brown out Reset
* @param  none
* @return none
*********************************************************************************/
void FeedTheDog()
{
  // feed dog 0
  TIMERG0.wdt_feed=1;                       // feed dog
  TIMERG0.wdt_wprotect=0;                   // write protect
  // feed dog 1
  TIMERG1.wdt_wprotect=TIMG_WDT_WKEY_VALUE; // write enable
  TIMERG1.wdt_feed=1;                       // feed dog
  TIMERG1.wdt_wprotect=0;                   // write protect
}

/*********************************************************************************
* @brief  readStringFile Read File from Flash Drive
* @param  none
* @return The Flash Text as String
*********************************************************************************/
String readStringFile(fs::FS &fs, const char * path)
{
    
    String Result;
    
    Serial.printf("Reading file: %s\r\n", path);

    File file = fs.open(path, "r" );
    if(!file || file.isDirectory()){
        Serial.println("- failed to open file for reading");
        return ("Error - failed to open file for reading");
    }

    Serial.println("- read from file:");

    while(file.available())
    {
    Result += char(file.read());
    }

file.close();
return (Result);
}

/*********************************************************************************
* @brief  writeFile 
* @param  message Text to write to Flash
* @return none
*********************************************************************************/
void writeFile(fs::FS &fs, const char * path, String message)
{
    Serial.printf("Writing file: %s\r\n", path);

    File file = fs.open(path, "w");
    if(!file){
        Serial.println("- failed to open file for writing");
        return;
    }
    if(file.println(message))
    {
        Serial.println("- file written");
    } else {
        Serial.println("- write failed");
    }
    file.close();
}


/*********************************************************************************
* @brief  GetEntries
* @param  filename of Ini File
* @return Integer of Entries 
*********************************************************************************/
int GetEntries(const char filename[])
{

  SPIFFSIniFile ini(filename);
  if (!ini.open()) {
    Serial.print("Ini file ");
    Serial.print(filename);
    Serial.println(" does not exist");
    // Cannot do anything else
    while (1);
  }
  Serial.println("Ini file exists");

  // Check the file is valid. This can be used to warn if any lines
  // are longer than the buffer.
  if (!ini.validate(Buffer, BufferLen)) {
    Serial.print("ini file ");
    Serial.print(ini.getFilename());
    Serial.print(" not valid: ");
    printErrorMessage(ini.getError());
    // Cannot do anything else
    while (1);
  }

int FileEntries = -1;

  if (ini.getValue("Entries", "Stations", Buffer, BufferLen, FileEntries)) {
       Serial.printf( "INI Stations has %d Entries",  FileEntries);
  }
  else {
    Serial.print("error was ");
    printErrorMessage(ini.getError());
  }
return (FileEntries);
}


/*********************************************************************************
* @brief  SaveRadioStation
* @param  FileName of Ini File
* @return none
*********************************************************************************/
void SaveRadioStation(const char FileName[])
{
writeFile(SPIFFS, FileName, MakeIniStations());
}



/*********************************************************************************
* @brief  printErrorMessage
* @param  e Error in INI File Access
* @return none
*********************************************************************************/
void printErrorMessage(uint8_t e, bool eol)
{
  switch (e) {
  case SPIFFSIniFile::errorNoError:
  #if UDP_PRINT2     
    udp_printf("no error");
  #else
 Serial.print("no error");
  #endif
  break;
  case SPIFFSIniFile::errorFileNotFound:
  #if UDP_PRINT2     
     udp_printf("file not found");
  #else
    Serial.print("file not found");
  #endif  
    break;
  case SPIFFSIniFile::errorFileNotOpen:
  #if UDP_PRINT2     
    udp_printf("file not open");
  #else
    Serial.print("file not open");
  #endif  
    break;
  case SPIFFSIniFile::errorBufferTooSmall:
  #if UDP_PRINT2     
    udp_printf("buffer too small");
  #else
    Serial.print("buffer too small");
  #endif  
    break;
  case SPIFFSIniFile::errorSeekError:
  #if UDP_PRINT2     
   udp_printf("seek error");
  #else
    Serial.print("seek error");
  #endif
    break;
  case SPIFFSIniFile::errorSectionNotFound:
  #if UDP_PRINT2     
    udp_printf("section not found");
  #else
    Serial.print("section not found");
  #endif
    break;
  case SPIFFSIniFile::errorKeyNotFound:
  #if UDP_PRINT2     
    udp_printf("key not found");
  #else
    Serial.print("key not found");
  #endif
    break;
  case SPIFFSIniFile::errorEndOfFile:
  #if UDP_PRINT2     
    udp_printf("end of file");
  #else
    Serial.print("end of file");
  #endif
    break;
  case SPIFFSIniFile::errorUnknownError:
  #if UDP_PRINT2     
    udp_printf("unknown error");
  #else
    Serial.print("unknown error");
  #endif
    break;
  default:
  #if UDP_PRINT2     
   udp_printf("unknown error value");
  #else
    Serial.print("unknown error value");
  #endif
    break;
  }
  if (eol)
    Serial.println();
}

/*********************************************************************************
* @brief  SetHTMLStation Prepare HTML for index.html
* @param  none
* @return none
*********************************************************************************/
void SetHTMLStation()
{
StationTabel.clear(); 

for (int i = 0; i < _Stations.Entries; i++ )
{
    Radio_t  Value = GetRadioStation(i);
    sprintf(Buffer, "<option value=\"%d\">%s</option> \n" ,i, Value.Name.c_str());
    #if UDP_PRINT 
     udp_printf("##############-->>>>>> %s \n", Buffer );
    #else 
     Serial.printf("##############-->>>>>> %s \n", Buffer );
  #endif 
  StationTabel =  StationTabel + Buffer + '\n';
}
  
   Serial.println("-------------------------------------------------") ;
    Serial.printf("Stationen : %s -- _Stations Entries : %d", StationTabel.c_str(),_Stations.Entries );
    Serial.println("-------------------------------------------------") ;
}

/*********************************************************************************
* @brief  SetupStations Prepare vetor array
* @param   filename in SPIFF
* @return none
*********************************************************************************/
void SetupStations(const char * filename)
{
  SPIFFSIniFile ini(filename);
  if (!ini.open()) {
  #if UDP_PRINT     
     udp_printf("Ini file ");
     udp_printf(filename);
     udp_printf(" does not exist \n");
  #else
      Serial.print("Ini file ");
      Serial.print(filename);
      Serial.println(" does not exist");
   #endif 
    // Cannot do anything else
    while (1);
  }
  Serial.println("Ini file exists");

  // Check the file is valid. This can be used to warn if any lines
  // are longer than the buffer.
  if (!ini.validate(Buffer, BufferLen)) {
    Serial.print("ini file ");
    Serial.print(ini.getFilename());
    Serial.print(" not valid: ");
    printErrorMessage(ini.getError());
    // Cannot do anything else
    while (1);
  }

char IniKey[10];

    _Stations.Entries = GetEntries(INI_FILENAME);

for (int i = 0 ; i < _Stations.Entries  ; i++)
{
sprintf_P(IniKey,"Url_%d", i);
Serial.println(IniKey);

if (ini.getValue("Station_URL", IniKey, Buffer, BufferLen)) {
if (strlen(Buffer) > 0 ){
   _Radio.url = Buffer;
    Serial.println(Buffer);
   
}
  } else {
    Serial.print("Error was ");
    printErrorMessage(ini.getError());
  }

//****************************************************************************************

sprintf_P(IniKey,"Name_%d", i);
Serial.println(IniKey);

if (ini.getValue("Station_Name", IniKey, Buffer, BufferLen)) {
if (strlen(Buffer) > 0 )
{
   _Radio.Name = Buffer;
    #if UDP_PRINT     
udp_printf(Buffer);
  #else
    Serial.println(Buffer);
  #endif  
//    sprintf(Buffer, "<option value=\"%d\">%s</option> \n" ,StationCounter++, _Station.Name.c_str() );

}
  }
  else {
    Serial.print("Error was ");
    printErrorMessage(ini.getError());
  }

// _Stations.url.clear();
// _Stations.Name.clear();

_Stations.url.push_back( _Radio.url );
_Stations.Name.push_back( _Radio.Name );

 #if UDP_PRINT     
udp_printf(" URL %s  \n", _Radio.url);
udp_printf(" URL %s  \n", _Radio.Name);      
#else      
Serial.println(_Radio.url);
Serial.println( _Radio.Name);
#endif
}
}

/*********************************************************************************
* @brief  AddNewStation
* @param  String url, String Name
* @return none
*********************************************************************************/
void AddNewStation(String url, String Name)
{
   _Stations.url.push_back(url);
   _Stations.Name.push_back(Name);
   _Stations.Entries = _Stations.Name.size();
}

/*********************************************************************************
* @brief  DeleteStation
* @param  int pos
* @return none
*********************************************************************************/
void DeleteStation(int pos)
{
   _Stations.url.erase(_Stations.url.begin() + pos);
   _Stations.Name.erase(_Stations.Name.begin() + pos);
   _Stations.Entries = _Stations.Name.size();
}

/*********************************************************************************
* @brief  ShowValue
* @param  int pos
* @return none
*********************************************************************************/
void ShowValue(int pos)
{
String S;

auto s = _Stations.url.begin() + pos;
      S = *s;
      Serial.printf("Value %d is %s \n", pos, S.c_str() );

  s = _Stations.Name.begin() + pos;

      S = *s;
        Serial.printf("Value %d is %s \n", pos, S.c_str() );
}

/*********************************************************************************
* @brief  GetRadioStation
* @param  int pos
* @return none
*********************************************************************************/
Radio_t GetRadioStation(int pos)
{
Radio_t Result;

auto s = _Stations.url.begin() + pos;
      Result.url = *s;
      //  printf("Value %d is %s \n", pos, S.c_str() );

  s = _Stations.Name.begin() + pos;

      Result.Name = *s;
        // printf("Value %d is %s \n", pos, S.c_str() );

return (Result);
}


/*********************************************************************************
* @brief  MakeIniStations
* @param  none
* @return none
*********************************************************************************/
String MakeIniStations()
{
String S;
Radio_t r;

S = "[Station_URL]\n";

for (int i = 0; i < _Stations.Entries; i++ )
{
r =  GetRadioStation(i);
sprintf(Buffer, "URL_%d = " , i );
S += Buffer + r.url + '\n'; 
}

S += "[Station_Name]\n";

for (int i = 0; i < _Stations.Entries; i++ )
{
r =  GetRadioStation(i);
sprintf(Buffer, "Name_%d = " , i );
S += Buffer + r.Name + '\n'; 

}

S += "[Entries]\n";
S += "Stations = ";
itoa(_Stations.Entries, Buffer ,10);
S += Buffer; 
return (S);

}

/*********************************************************************************
* @brief  SendUDPPackets
* @param  Value
* @return none
*********************************************************************************/
void SendUDPPackets(const char Value[] )
{ 
  Udp.beginPacket(UdpAddress, UdpPort);
  Udp.print(Value);
  Udp.endPacket();
  // memset(Buffer, 0, BufferLen );

}

/*********************************************************************************
* @brief  ReceiveUdpPacket not used
* @param  Value
* @return none
*********************************************************************************/
void ReceiveUdpPacket()
{
    // You must use this in Loop() to recive UDP Packets
  //udp.parsePacket();
  if(Udp.read(Buffer, 50) > 0){
    Serial.print("UDP--> ");
    Serial.println((char *)Buffer);
  }
}


/*********************************************************************************
* @brief  udp_printf
* @param  const char * format, ... 
* @return none
*********************************************************************************/
void udp_printf ( const char * format, ... )
{
  va_list args;
  va_start (args, format);
  vsprintf (Buffer, format, args);

         Udp.beginPacket(UdpAddress, UdpPort);
        Udp.print(Buffer);
        Udp.endPacket();

  va_end (args);
}


/*********************************************************************************
* @brief  ReadFile  to read the Content of the SPIFF Text File
* @param  fs::FS &fs, const char * path
* @return none
*********************************************************************************/
void ReadFile(fs::FS &fs, const char * path)
{
    Serial.printf("Reading file: %s\r\n", path);

    File file = fs.open(path);
    if(!file || file.isDirectory()){
        Serial.println("- failed to open file for reading");
        return;
    }

    Serial.println("- read from file:");
    while(file.available()){
        Serial.write(file.read());
    }
    file.close();
}



